package java.util.jar;

import java.security.CodeSigner;
import java.security.cert.Certificate;
import java.util.Iterator;
import sun.misc.SharedSecrets;
import java.util.ArrayList;
import java.util.List;
import java.net.URL;
import java.util.NoSuchElementException;
import java.security.CodeSource;
import java.security.AccessController;
import sun.security.action.GetPropertyAction;
import java.io.InputStream;
import sun.misc.IOUtils;
import sun.security.util.ManifestEntryVerifier;
import sun.security.util.SignatureFileVerifier;
import java.util.Locale;
import java.util.stream.StreamSupport;
import java.util.Spliterators;
import java.util.stream.Stream;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.File;
import java.lang.ref.SoftReference;
import java.util.zip.ZipFile;

public class JarFile extends ZipFile
{
    private SoftReference<Manifest> manRef;
    private JarEntry manEntry;
    private JarVerifier jv;
    private boolean jvInitialized;
    private boolean verify;
    static final ThreadLocal<Boolean> isInitializing;
    private boolean hasClassPathAttribute;
    private volatile boolean hasCheckedSpecialAttributes;
    public static final String MANIFEST_NAME = "META-INF/MANIFEST.MF";
    private static final char[] CLASSPATH_CHARS;
    private static final int[] CLASSPATH_LASTOCC;
    private static final int[] CLASSPATH_OPTOSFT;
    private static String javaHome;
    private static volatile String[] jarNames;
    
    public JarFile(String pathname) throws IOException {
        this(new File(pathname), true, 1);
    }
    
    public JarFile(String pathname, boolean b) throws IOException {
        this(new File(pathname), b, 1);
    }
    
    public JarFile(File file) throws IOException {
        this(file, true, 1);
    }
    
    public JarFile(File file, boolean b) throws IOException {
        this(file, b, 1);
    }
    
    public JarFile(File file, boolean verify, int mode) throws IOException {
        super(file, mode);
        this.verify = verify;
    }
    
    public Manifest getManifest() throws IOException {
        return this.getManifestFromReference();
    }
    
    private Manifest getManifestFromReference() throws IOException {
        Manifest referent = (this.manRef != null) ? ((Manifest)this.manRef.get()) : null;
        if (referent == null) {
            JarEntry manEntry = this.getManEntry();
            if (manEntry != null) {
                if (this.verify) {
                    byte[] bytes = this.getBytes(manEntry);
                    if (!this.jvInitialized) {
                        this.jv = new JarVerifier(bytes);
                    }
                    referent = new Manifest(this.jv, new ByteArrayInputStream(bytes));
                }
                else {
                    referent = new Manifest(super.getInputStream(manEntry));
                }
                this.manRef = new SoftReference(referent);
            }
        }
        return referent;
    }
    
    private native String[] getMetaInfEntryNames();
    
    public JarEntry getJarEntry(String s) {
        return (JarEntry)this.getEntry(s);
    }
    
    @Override
    public ZipEntry getEntry(String name) {
        ZipEntry entry = super.getEntry(name);
        if (entry != null) {
            return new JarFile.JarFileEntry(entry);
        }
        return null;
    }
    
    @Override
    public Enumeration<JarEntry> entries() {
        return new JarFile.JarEntryIterator();
    }
    
    @Override
    public Stream<JarEntry> stream() {
        return StreamSupport.stream(Spliterators.spliterator(new JarFile.JarEntryIterator(), (long)this.size(), 1297), false);
    }
    
    private void maybeInstantiateVerifier() throws IOException {
        if (this.jv != null) {
            return;
        }
        if (this.verify) {
            String[] metaInfEntryNames = this.getMetaInfEntryNames();
            if (metaInfEntryNames != null) {
                int i = 0;
                while (i < metaInfEntryNames.length) {
                    String upperCase = metaInfEntryNames[i].toUpperCase(Locale.ENGLISH);
                    if (upperCase.endsWith(".DSA") || upperCase.endsWith(".RSA") || upperCase.endsWith(".EC") || upperCase.endsWith(".SF")) {
                        this.getManifest();
                        return;
                    }
                    ++i;
                }
            }
            this.verify = false;
        }
        return;
    }
    
    private void initializeVerifier() {
        Object o = null;
        try {
            String[] metaInfEntryNames = this.getMetaInfEntryNames();
            if (metaInfEntryNames != null) {
                for (int i = 0; i < metaInfEntryNames.length; ++i) {
                    String upperCase = metaInfEntryNames[i].toUpperCase(Locale.ENGLISH);
                    if ("META-INF/MANIFEST.MF".equals(upperCase) || SignatureFileVerifier.isBlockOrSF(upperCase)) {
                        JarEntry jarEntry = this.getJarEntry(metaInfEntryNames[i]);
                        if (jarEntry == null) {
                            throw new JarException("corrupted jar file");
                        }
                        if (o == null) {
                            o = new ManifestEntryVerifier(this.getManifestFromReference());
                        }
                        byte[] bytes = this.getBytes(jarEntry);
                        if (bytes != null && bytes.length > 0) {
                            this.jv.beginEntry(jarEntry, o);
                            this.jv.update(bytes.length, bytes, 0, bytes.length, o);
                            this.jv.update(-1, null, 0, 0, o);
                        }
                    }
                }
            }
        }
        catch (IOException ex) {
            this.jv = null;
            this.verify = false;
            if (JarVerifier.debug != null) {
                JarVerifier.debug.println("jarfile parsing error!");
                ex.printStackTrace();
            }
        }
        if (this.jv != null) {
            this.jv.doneWithMeta();
            if (JarVerifier.debug != null) {
                JarVerifier.debug.println("done with meta!");
            }
            if (this.jv.nothingToVerify()) {
                if (JarVerifier.debug != null) {
                    JarVerifier.debug.println("nothing to verify!");
                }
                this.jv = null;
                this.verify = false;
            }
        }
    }
    
    private byte[] getBytes(ZipEntry entry) throws IOException {
        try (final InputStream inputStream = super.getInputStream(entry)) {
            return IOUtils.readFully(inputStream, (int)entry.getSize(), true);
        }
    }
    
    @Override
    public synchronized InputStream getInputStream(ZipEntry entry) throws IOException {
        this.maybeInstantiateVerifier();
        if (this.jv == null) {
            return super.getInputStream(entry);
        }
        if (!this.jvInitialized) {
            this.initializeVerifier();
            this.jvInitialized = true;
            if (this.jv == null) {
                return super.getInputStream(entry);
            }
        }
        return new JarVerifier.VerifierStream(this.getManifestFromReference(), (entry instanceof JarFile.JarFileEntry) ? ((JarEntry)entry) : this.getJarEntry(entry.getName()), super.getInputStream(entry), this.jv);
    }
    
    private JarEntry getManEntry() {
        if (this.manEntry == null) {
            this.manEntry = this.getJarEntry("META-INF/MANIFEST.MF");
            if (this.manEntry == null) {
                String[] metaInfEntryNames = this.getMetaInfEntryNames();
                if (metaInfEntryNames != null) {
                    int i = 0;
                    while (i < metaInfEntryNames.length) {
                        if ("META-INF/MANIFEST.MF".equals(metaInfEntryNames[i].toUpperCase(Locale.ENGLISH))) {
                            this.manEntry = this.getJarEntry(metaInfEntryNames[i]);
                            break;
                        }
                        ++i;
                    }
                }
            }
        }
        return this.manEntry;
    }
    
    boolean hasClassPathAttribute() throws IOException {
        this.checkForSpecialAttributes();
        return this.hasClassPathAttribute;
    }
    
    private boolean match(char[] array, byte[] array2, int[] array3, int[] array4) {
        int length = array.length;
        int n = array2.length - length;
        int i = 0;
    Label_0014:
        while (i <= n) {
            int j = length - 1;
            while (j >= 0) {
                int n2 = (char)array2[i + j];
                int n3 = ((n2 - 65 | 90 - n2) >= 0) ? ((char)(n2 + 32)) : n2;
                if (n3 != array[j]) {
                    i += Math.max(j + 1 - array3[n3 & '\u007f'], array4[j]);
                    continue Label_0014;
                }
                --j;
            }
            return true;
        }
        return false;
    }
    
    private void checkForSpecialAttributes() throws IOException {
        if (this.hasCheckedSpecialAttributes) {
            return;
        }
        if (!this.isKnownNotToHaveSpecialAttributes()) {
            JarEntry manEntry = this.getManEntry();
            if (manEntry != null && this.match(JarFile.CLASSPATH_CHARS, this.getBytes(manEntry), JarFile.CLASSPATH_LASTOCC, JarFile.CLASSPATH_OPTOSFT)) {
                this.hasClassPathAttribute = true;
            }
        }
        this.hasCheckedSpecialAttributes = true;
        return;
    }
    
    private boolean isKnownNotToHaveSpecialAttributes() {
        if (JarFile.javaHome == null) {
            JarFile.javaHome = (String)AccessController.doPrivileged(new GetPropertyAction("java.home"));
        }
        if (JarFile.jarNames == null) {
            String[] jarNames = new String[11];
            String separator = File.separator;
            int n = 0;
            jarNames[n++] = separator + "rt.jar";
            jarNames[n++] = separator + "jsse.jar";
            jarNames[n++] = separator + "jce.jar";
            jarNames[n++] = separator + "charsets.jar";
            jarNames[n++] = separator + "dnsns.jar";
            jarNames[n++] = separator + "zipfs.jar";
            jarNames[n++] = separator + "localedata.jar";
            String str = jarNames[n++] = "cldrdata.jar";
            jarNames[n++] = str + "sunjce_provider.jar";
            jarNames[n++] = str + "sunpkcs11.jar";
            jarNames[n++] = str + "sunec.jar";
            JarFile.jarNames = jarNames;
        }
        String name = this.getName();
        if (name.startsWith(JarFile.javaHome)) {
            String[] jarNames2 = JarFile.jarNames;
            int i = 0;
            while (i < jarNames2.length) {
                if (name.endsWith(jarNames2[i])) {
                    return true;
                }
                ++i;
            }
        }
        return false;
    }
    
    synchronized void ensureInitialization() {
        try {
            this.maybeInstantiateVerifier();
        }
        catch (IOException cause) {
            throw new RuntimeException(cause);
        }
        if (this.jv != null && !this.jvInitialized) {
            JarFile.isInitializing.set(Boolean.TRUE);
            try {
                this.initializeVerifier();
                this.jvInitialized = true;
            }
            finally {
                JarFile.isInitializing.set(Boolean.FALSE);
            }
        }
    }
    
    JarEntry newEntry(ZipEntry zipEntry) {
        return new JarFile.JarFileEntry(zipEntry);
    }
    
    Enumeration<String> entryNames(CodeSource[] array) {
        this.ensureInitialization();
        if (this.jv != null) {
            return this.jv.entryNames(this, array);
        }
        boolean n = false;
        int i = 0;
        while (i < array.length) {
            if (array[i].getCodeSigners() == null) {
                n = true;
                break;
            }
            ++i;
        }
        if (!(n == 0)) {
            return this.unsignedEntryNames();
        }
        return new Enumeration<String>() {
            @Override
            public boolean hasMoreElements() {
                return false;
            }
            
            @Override
            public String nextElement() {
                throw new NoSuchElementException();
            }
        };
    }
    
    Enumeration<JarEntry> entries2() {
        this.ensureInitialization();
        if (this.jv != null) {
            return this.jv.entries2(this, super.entries());
        }
        return new Enumeration<JarEntry>() {
            ZipEntry entry;
            
            @Override
            public boolean hasMoreElements() {
                if (this.entry != null) {
                    return true;
                }
                while (enum_.hasMoreElements()) {
                    ZipEntry entry = (ZipEntry)enum_.nextElement();
                    if (JarVerifier.isSigningRelated(entry.getName())) {
                        continue;
                    }
                    this.entry = entry;
                    return true;
                }
                return false;
            }
            
            @Override
            public JarFile.JarFileEntry nextElement() {
                if (this.hasMoreElements()) {
                    ZipEntry entry = this.entry;
                    this.entry = null;
                    return new JarFile.JarFileEntry(entry);
                }
                throw new NoSuchElementException();
            }
        };
    }
    
    CodeSource[] getCodeSources(URL url) {
        this.ensureInitialization();
        if (this.jv != null) {
            return this.jv.getCodeSources(this, url);
        }
        if (this.unsignedEntryNames().hasMoreElements()) {
            return new CodeSource[] { JarVerifier.getUnsignedCS(url) };
        }
        return null;
    }
    
    private Enumeration<String> unsignedEntryNames() {
        return new Enumeration<String>() {
            String name;
            
            @Override
            public boolean hasMoreElements() {
                if (this.name != null) {
                    return true;
                }
                while (entries.hasMoreElements()) {
                    ZipEntry zipEntry = (ZipEntry)entries.nextElement();
                    String name = zipEntry.getName();
                    if (zipEntry.isDirectory()) {
                        continue;
                    }
                    if (JarVerifier.isSigningRelated(name)) {
                        continue;
                    }
                    this.name = name;
                    return true;
                }
                return false;
            }
            
            @Override
            public String nextElement() {
                if (this.hasMoreElements()) {
                    String name = this.name;
                    this.name = null;
                    return name;
                }
                throw new NoSuchElementException();
            }
        };
    }
    
    CodeSource getCodeSource(URL url, String s) {
        this.ensureInitialization();
        if (this.jv != null) {
            if (this.jv.eagerValidation) {
                JarEntry jarEntry = this.getJarEntry(s);
                if (jarEntry != null) {
                    return this.jv.getCodeSource(url, this, jarEntry);
                }
                return this.jv.getCodeSource(url, s);
            }
            else {
                return this.jv.getCodeSource(url, s);
            }
        }
        else {
            return JarVerifier.getUnsignedCS(url);
        }
    }
    
    void setEagerValidation(boolean eagerValidation) {
        try {
            this.maybeInstantiateVerifier();
        }
        catch (IOException cause) {
            throw new RuntimeException(cause);
        }
        if (this.jv != null) {
            this.jv.setEagerValidation(eagerValidation);
        }
    }
    
    List<Object> getManifestDigests() {
        this.ensureInitialization();
        if (this.jv != null) {
            return this.jv.getManifestDigests();
        }
        return new ArrayList();
    }
    
    static boolean isInitializing() {
        Boolean b = (Boolean)JarFile.isInitializing.get();
        return b != null && (boolean)b;
    }
    
    static Enumeration access$001(JarFile jarFile) {
        return jarFile.entries();
    }
    
    static void access$200(JarFile jarFile) throws IOException {
        jarFile.maybeInstantiateVerifier();
    }
    
    static {
        isInitializing = new ThreadLocal();
        SharedSecrets.setJavaUtilJarAccess(new JavaUtilJarAccessImpl());
        CLASSPATH_CHARS = new char[] { 'c', 'l', 'a', 's', 's', '-', 'p', 'a', 't', 'h' };
        CLASSPATH_LASTOCC = new int[128];
        CLASSPATH_OPTOSFT = new int[10];
        JarFile.CLASSPATH_LASTOCC[99] = 1;
        JarFile.CLASSPATH_LASTOCC[108] = 2;
        JarFile.CLASSPATH_LASTOCC[115] = 5;
        JarFile.CLASSPATH_LASTOCC[45] = 6;
        JarFile.CLASSPATH_LASTOCC[112] = 7;
        JarFile.CLASSPATH_LASTOCC[97] = 8;
        JarFile.CLASSPATH_LASTOCC[116] = 9;
        JarFile.CLASSPATH_LASTOCC[104] = 10;
        for (int i = 0; i < 9; ++i) {
            JarFile.CLASSPATH_OPTOSFT[i] = 10;
        }
        JarFile.CLASSPATH_OPTOSFT[9] = 1;
    }
    
    private class JarEntryIterator implements Enumeration<JarEntry>, Iterator<JarEntry>
    {
        final Enumeration<? extends ZipEntry> e;
        
        private JarEntryIterator() {
            this.e = JarFile.access$001(JarFile.this);
        }
        
        @Override
        public boolean hasNext() {
            return this.e.hasMoreElements();
        }
        
        @Override
        public JarEntry next() {
            return new JarFile.JarFileEntry((ZipEntry)this.e.nextElement());
        }
        
        @Override
        public boolean hasMoreElements() {
            return this.hasNext();
        }
        
        @Override
        public JarEntry nextElement() {
            return this.next();
        }
    }
    
    private class JarFileEntry extends JarEntry
    {
        JarFileEntry(ZipEntry zipEntry) {
            super(zipEntry);
        }
        
        @Override
        public Attributes getAttributes() throws IOException {
            Manifest manifest = JarFile.this.getManifest();
            if (manifest != null) {
                return manifest.getAttributes(this.getName());
            }
            return null;
        }
        
        @Override
        public Certificate[] getCertificates() {
            try {
                JarFile.access$200(JarFile.this);
            }
            catch (IOException cause) {
                throw new RuntimeException(cause);
            }
            if (this.certs == null && JarFile.this.jv != null) {
                this.certs = JarFile.this.jv.getCerts(JarFile.this, this);
            }
            return (this.certs == null) ? null : ((Certificate[])this.certs.clone());
        }
        
        @Override
        public CodeSigner[] getCodeSigners() {
            try {
                JarFile.access$200(JarFile.this);
            }
            catch (IOException cause) {
                throw new RuntimeException(cause);
            }
            if (this.signers == null && JarFile.this.jv != null) {
                this.signers = JarFile.this.jv.getCodeSigners(JarFile.this, this);
            }
            return (this.signers == null) ? null : ((CodeSigner[])this.signers.clone());
        }
    }
}
